
Require Import Compare_dec.
Require Import Min.
Require Import Max.
Require Import Le.
Require Import Omega.
Require abstraction.

Section Natural_inf_domain.
Lemma natural_inf_domain_union_correct1 :
    forall a b : nat,
    forall x : nat,
    (le a x) -> (le (min a b) x).
Proof.
  intros.
  apply le_trans with (m := a).
   apply le_min_l.
   assumption.
Qed.

Lemma natural_inf_domain_union_correct2 :
    forall a b : nat,
    forall x : nat,
    (le b x) -> (le (min a b) x).
Proof.
  intros.
  apply le_trans with (m := b).
   apply le_min_r.
   assumption.
Qed.

Lemma natural_inf_domain_le_correct:
    forall xAbstr yAbstr : nat,
    (ge xAbstr yAbstr) ->
    forall x : nat,
    (le xAbstr x) -> (le yAbstr x).
Proof.
  intros.
  omega.
Qed.

Lemma natural_inf_domain_le_refl : forall x : nat, (ge x x).
Proof.
  auto with arith.
Qed.

Lemma natural_inf_domain_le_trans : forall x y z : nat,
    (ge x y) -> (ge y z) -> (ge x z).
Proof.
  intros.
  omega.
Qed.

Definition  natural_inf_domain_widening_chain : (abstraction.widening_chain nat ge).
Proof.
  apply abstraction.widening_chain_intro with (x := 0%nat).
  intros.
  right.
   auto with arith.
Defined.

Definition natural_inf_domain_le_decide :
  forall x y : nat, { ge x y } + { ~(ge x y) }.
Proof.
  intros.
  destruct (le_gt_dec y x).
   left.
     omega.
   right.
     omega.
Qed.

Definition natural_inf_domain_iter_union :
    forall x : nat,
    forall y : nat,
    { z : nat | ge x z } + { ge y x }.
Proof.
  intros.
  destruct (le_ge_dec x y).
   right; assumption.
   left.
     exists y.
     assumption.
Qed.

Inductive natural_inf_domain_zeroary_ops : Set := .

Definition natural_inf_domain_zeroary_predicate
  (op : natural_inf_domain_zeroary_ops) (m : nat): Prop :=
  match op with
  (* | natural_inf_domain_inter => (m1 <= m3) /\ (m2 <= m3) *)
  end.

Definition natural_inf_domain_zeroary (op : natural_inf_domain_zeroary_ops) : nat :=
  match op with
  end.

Lemma natural_inf_domain_zeroary_correct :
    forall op : natural_inf_domain_zeroary_ops,
    forall x : nat,
    (natural_inf_domain_zeroary_predicate op x) ->
    (le (natural_inf_domain_zeroary op) x).
Proof.
  intros.
  destruct op.
Qed.

Inductive natural_inf_domain_unary_ops : Set :=
  | natural_inf_domain_F :  natural_inf_domain_unary_ops.

Definition natural_inf_domain_unary_predicate
  (op : natural_inf_domain_unary_ops)
  (m1 : nat) (m2 : nat) : Prop :=
  match op with
  | natural_inf_domain_F => (m2 >= 4) /\ (m2 = m1 - 1)
  end.

Definition natural_inf_domain_unary (op : natural_inf_domain_unary_ops)
  (m : nat) : nat :=
  match op with
  | natural_inf_domain_F => max 4%nat (minus m 1%nat)
  end.

Lemma natural_inf_domain_unary_correct :
    forall op : natural_inf_domain_unary_ops,
    forall a : nat,
    forall x y : nat,
    (le a x) ->
    (natural_inf_domain_unary_predicate op x y) ->
    (le (natural_inf_domain_unary op a) y).
Proof.
  unfold natural_inf_domain_unary_predicate, natural_inf_domain_unary in |- *.
  destruct op.
  intros until y.
  intro Habstract.
  intro H; destruct H.
  destruct (max_dec 4 (a - 1)) as [e| e].
   rewrite e.
     auto with arith.
   rewrite e.
     omega.
Qed.

Inductive natural_inf_domain_binary_ops : Set := (*
  | natural_inf_domain_inter :  natural_inf_domain_binary_ops *).

Definition natural_inf_domain_binary_predicate
  (op : natural_inf_domain_binary_ops)
  (m1 : nat) (m2 : nat) (m3 : nat) : Prop :=
  match op with
  (* | natural_inf_domain_inter => (m1 <= m3) /\ (m2 <= m3) *)
  end.

Definition natural_inf_domain_binary (op : natural_inf_domain_binary_ops)
  (m1 : nat) (m2 : nat) : nat :=
  match op with
  (* | natural_inf_domain_inter => max m1 m2 *)
  end.

Lemma natural_inf_domain_binary_correct :
    forall op : natural_inf_domain_binary_ops,
    forall a b : nat,
    forall x y z : nat,
    (le a x) ->
    (le b y) ->
    (natural_inf_domain_binary_predicate op x y z) ->
    (le (natural_inf_domain_binary op a b) z).
Proof.
  intros.
  destruct op.
Qed.


Definition natural_inf_domain :=
  (abstraction.Build_domain nat (* concrete *)
   natural_inf_domain_zeroary_ops
   natural_inf_domain_zeroary_predicate
   natural_inf_domain_unary_ops
   natural_inf_domain_unary_predicate
   natural_inf_domain_binary_ops
   natural_inf_domain_binary_predicate

   nat (* domain *)
   le (* x abstracts y <=> x <= y *)

   natural_inf_domain_zeroary
   natural_inf_domain_zeroary_correct
   natural_inf_domain_unary
   natural_inf_domain_unary_correct
   natural_inf_domain_binary
   natural_inf_domain_binary_correct

   min (* union *)
   natural_inf_domain_union_correct1 (* union correct wrt left *)
   natural_inf_domain_union_correct2 (* union correct wrt right *)

   ge (* x abstracts y <=> x <= y *)
   natural_inf_domain_le_refl
   natural_inf_domain_le_trans
   natural_inf_domain_le_decide
   natural_inf_domain_le_correct

   natural_inf_domain_iter_union
   natural_inf_domain_widening_chain).
End  Natural_inf_domain.

Require Import List.

Definition test_domain_ramp1 := abstraction.add_widening_ramp
  natural_inf_domain
    (100 :: 90 :: 80 :: 70 :: 10 :: 5 :: 2 :: 1 :: nil)%nat.
Definition test_domain_ramp2 := abstraction.add_widening_ramp
  natural_inf_domain
    (100 :: 95 :: 85 :: 60 :: 7 :: 3 :: 1 :: nil)%nat.

Definition test_domain1 :=
  (abstraction.delay_widening_each_step test_domain_ramp1 2).
Definition test_domain2 :=
  (abstraction.delay_widening_each_step test_domain_ramp2 1).

Definition test_domain := (abstraction.product_domain
   test_domain1 test_domain2).

Axiom logf1 : nat -> nat.
Definition test_f1 (m : nat) :=
  abstraction.domain_unary test_domain1 natural_inf_domain_F (logf1 m).
Axiom logf2 : nat -> nat.
Definition test_f2 (m : nat) :=
  abstraction.domain_unary test_domain2 natural_inf_domain_F (logf2 m).
Definition test_f (x12 : (prod nat nat)) :=
  ((test_f1 (fst x12)), (test_f2 (snd x12))).

Definition final1 : nat := (proj1_sig (abstraction.abstract_lfp
  test_domain1 test_f1)).
Definition final : (prod nat nat) := (proj1_sig (abstraction.abstract_lfp
  test_domain test_f)).

Extract Inlined Constant logf1 => "logf1".
Extract Inlined Constant logf2 => "logf2".

(* nat => int64 *)
Extract Inlined Constant le_lt_dec => "(( <= ) :  int64->int64->bool)".
Extract Inlined Constant min => "(min : int64->int64->int64)".
Extract Inlined Constant max => "(max : int64->int64->int64)".
Extract Inlined Constant plus => "Int64.add".
Extract Inlined Constant minus => "Int64.sub".
Extract Inductive nat => "Int64.t" [ "Int64.zero" "Int64.succ"].
Extract Inlined Constant abstraction.functional_power => "functional_power".

(* basic types *)
Extract Inductive unit => "unit" [ "()" ].
Extract Inductive bool => "bool" [ "true" "false" ].
Extract Inductive sumbool => "bool" [ "true" "false" ].
Extract Inductive option => "option" [ "Some" "None"].
Extract Inductive prod => "pair" [ "" ].
Extract Inlined Constant fst => "fst".
Extract Inlined Constant snd => "snd".

Extraction Inline proj1_sig.
Extraction Inline le_gt_dec.

Extraction "test_coq.ml" final.
